@@ -1,14 +1,19 @@
 import pandas as pd
 from sqlalchemy import create_engine
+from dask import delayed
 
 class da_tran_SQL :
     """interact with SQL"""
-    def __init__(self, sql_type, host_name, database_name, user, password , chunksize = 150,  **kwargs):
+    def __init__(self, sql_type, host_name, database_name, user, password 
+                , chunksize = 150, partition_size = 5000, parallel_dump = False, max_parallel = 2, **kwargs):
         """Create connection to SQL Server"""
-
         sql_type = sql_type.upper()
         self.sql_type = sql_type
-        self.chunksize = chunksize
+        self.chunksize = int(chunksize)
+        self.partition_size = int(partition_size)
+        self.parallel_dump = parallel_dump
+        self.max_parallel = int(max_parallel)
+
         type_dic = {
                     'MSSQL' : ['mssql', 'pymssql', '1433','[',']','EXEC'],
                     'MYSQL' : ['mysql', 'pymysql','3306','','','CALL'],
@@ -25,6 +30,28 @@ class da_tran_SQL :
         self.engine = create_engine(connection_str)
         print(pd.read_sql_query("""SELECT 'Connection OK'""", con = self.engine).iloc[0,0]) 
 
+    def sub_dump(self, df_in,table_name_in,mode_in) :
+        """normal to_sql for dask's delayed in dump_main"""
+        df_in.to_sql(table_name_in, con = self.engine, index = False,if_exists = mode_in,chunksize = self.chunksize, method = 'multi')
+        return len(df_in)
+
+    def dump_main(self, df_in, table_name_in ,mode_in) :
+        """Divide and Dump Dataframe"""
+        if len(df_in) <= self.partition_size : sum_len = self.sub_dump(df_in,table_name_in,mode_in) 
+        else :
+            if self.parallel_dump : dask_dump = delayed(self.sub_dump)
+            i, j, sum_len, df_length = 0, 1, 0, len(df_in)
+            while i < df_length :
+                i_next = i + self.partition_size
+                if self.parallel_dump :
+                    sum_len += dask_dump(df_in.iloc[i:i_next,:],table_name_in,mode_in) 
+                    if (j == self.max_parallel) | (i_next >= df_length) : sum_len, j = sum_len.compute(), 1
+                else : sum_len += self.sub_dump(df_in.iloc[i:i_next,:],table_name_in,mode_in) 
+                j += 1
+                i += self.partition_size
+        return sum_len
+
+
     def read(self, table_name_in, condition_in = '', SP = False, param = ''):
         """Read Table or View or Store Procedure"""
         if SP :
@@ -56,35 +83,44 @@ class da_tran_SQL :
 
     def write_logic_sql(self, key , value_in):
         """Write SQL Condition Query that include >, <, ="""
-        logic_query = self.begin_name+ str(key) + self.end_name + ' ' + value_in['logic'] + ' '
+        #logic_query = self.begin_name+ str(key) + self.end_name + ' ' + value_in['logic'] + ' '
         value_in['type'] = value_in.get('type','')
         if value_in.get('value','there is no value') != 'there is no value' :
             if ('date' in value_in['type']) | ('time' in value_in['type']):  # Datetime will need '' in SQL Statement
+                logic_query = 'CAST(' + self.begin_name+ str(key) + self.end_name + ' AS {}) '.format(value_in['type']) + value_in['logic'] + ' '
                 logic_query += """'""" + str(value_in['value']) + """'"""
-            else : logic_query += str(value_in['value'])
+            else : 
+                logic_query = self.begin_name+ str(key) + self.end_name + ' ' + value_in['logic'] + ' ' + str(value_in['value'])
+#                logic_query += 
         else : raise Exception("Please insert value data")  # will get error if not specific value
         return logic_query
 
     def dump_whole(self, df_in, table_name_in , fix_table = False, debug = False) :
         """Delete exists table and replace with new df"""
         if fix_table : 
-            print('Start Filter Existing data from df at ',pd.Timestamp.now())
-            self.engine.execute("""DELETE FROM [{}]""".format(table_name_in))
+            print('Delete Existing data from Table at ',pd.Timestamp.now())
+            self.engine.execute("""DELETE FROM {}{}{}""".format(self.begin_name,table_name_in,self.end_name))
             #Dump df_in to database
-            df_in.to_sql(table_name_in,con = self.engine,index = False,if_exists = 'append',chunksize = self.chunksize, method = 'multi')
+            #df_in.to_sql(table_name_in,con = self.engine,index = False,if_exists = 'append',chunksize = self.chunksize, method = 'multi')
+            self.dump_main( df_in, table_name_in ,mode_in = 'append') 
             print('Dump data to ',table_name_in,' End ',pd.Timestamp.now())
         else :
-            print('Start Filter Existing data from df at ',pd.Timestamp.now())
+            print('Drop Existing Table at ',pd.Timestamp.now())
+            self.engine.execute("""DROP TABLE {}{}{}""".format(self.begin_name,table_name_in,self.end_name))
             #Dump df_in to database
-            df_in.to_sql(table_name_in,con = self.engine,index = False,if_exists = 'replace',chunksize = self.chunksize, method = 'multi')
+            #df_in.to_sql(table_name_in,con = self.engine,index = False,if_exists = 'replace',chunksize = self.chunksize, method = 'multi')
+            self.dump_main( df_in, table_name_in ,mode_in = 'append') 
             print('Dump data to ',table_name_in,' End ',pd.Timestamp.now())
 
     def write_in_sql(self,df_in , key) :
         """Write SQL Condition Query 'in (x,x,x)'"""
         filter_filter = tuple(df_in[key].fillna('Will BE rEpLaCe wItH NULL').astype('str').unique())
         if len(filter_filter) == 1 : 
-            filter_filter = str(filter_filter).replace(',)',')')  # tuple with 1 value will be ( x , ) => need to convert
-            logic_query = self.begin_name + key  + self.end_name + ' in ' + filter_filter
+            if 'Will BE rEpLaCe wItH NULL' in filter_filter :
+                logic_query = '({}{}{} IS NULL'.format(self.begin_name , key, self.end_name) + ')'
+            else :
+                filter_filter = str(filter_filter).replace(',)',')')  # tuple with 1 value will be ( x , ) => need to convert
+                logic_query = self.begin_name + key  + self.end_name + ' in ' + filter_filter
         elif len(filter_filter) > 1 :
             filter_filter = str(filter_filter) # tuple with > 1 values will be ( x, y, z) which can be use in SQL
             logic_query = self.begin_name + key  + self.end_name + ' in ' + filter_filter
@@ -94,6 +130,8 @@ class da_tran_SQL :
             #print(logic_query)
             logic_query = logic_query.replace(", 'Will BE rEpLaCe wItH NULL'",'')
             #print(logic_query)
+            logic_query = logic_query.replace("'Will BE rEpLaCe wItH NULL',",'')
+            logic_query = logic_query.replace("'Will BE rEpLaCe wItH NULL'",'')
             logic_query = '(' + logic_query + ' OR {}{}{} IS NULL'.format(self.begin_name , key, self.end_name) + ')'
 
         return logic_query
@@ -145,7 +183,8 @@ class da_tran_SQL :
             
         if debug : print(sql_q)
         #Dump df_in append to database
-        df_in.to_sql(table_name_in,con = self.engine,index = False,if_exists = 'append',chunksize = self.chunksize, method = 'multi')
+        #df_in.to_sql(table_name_in,con = self.engine,index = False,if_exists = 'append',chunksize = self.chunksize, method = 'multi')
+        self.dump_main(df_in, table_name_in ,mode_in = 'append')
         print('Dump data to ',table_name_in,' End ',pd.Timestamp.now())
 
     def dump_new(self, df_in, table_name_in, list_key , debug = False) :
@@ -193,7 +232,8 @@ class da_tran_SQL :
         df_in = df_in[logic_filter]
 
         #Dump df_in append to database
-        df_in.to_sql(table_name_in,con = self.engine,index = False,if_exists = 'append',chunksize = self.chunksize, method = 'multi')
+        #df_in.to_sql(table_name_in,con = self.engine,index = False,if_exists = 'append',chunksize = self.chunksize, method = 'multi')
+        self.dump_main( df_in, table_name_in ,mode_in = 'append') 
         print('Dump data to ',table_name_in,' End ', pd.Timestamp.now())
 
     def dump_new_old(self, df_in, table_name_in, list_key) :
@@ -230,5 +270,6 @@ class da_tran_SQL :
         df_in = df_in[logic_filter]
 
         #Dump df_in append to database
-        df_in.to_sql(table_name_in,con = self.engine,index = False,if_exists = 'append',chunksize = self.chunksize, method = 'multi')
+        #df_in.to_sql(table_name_in,con = self.engine,index = False,if_exists = 'append',chunksize = self.chunksize, method = 'multi')
+        self.dump_main(df_in, table_name_in ,mode_in = 'append')
         print('Dump data to ',table_name_in,' End ', pd.Timestamp.now())
\ No newline at end of file
